#include "dependency.h"

#include <regex>

#include "project.h"

constexpr std::string_view acceptJson("accept: application/json");
const std::regex http("^https?:\\/\\/");
const std::regex repo("(^https?):\\/\\/([\\w\\d\\.\\-\\_]+)\\/([\\w\\d\\-\\_]+)\\/([\\w\\d\\-\\_]+)");

Dependency::Dependency(
    const std::string& path,
    Type type,
    const std::optional<std::string>& name,
    const std::optional<std::string>& version
):
    path(path),
    type(type),
    name(name),
    version(version)
{}

Dependency::Type Dependency::getType() const {
    return type;
}

std::optional<std::string> Dependency::getName() const {
    return name;
}

std::optional<std::string> Dependency::getVersion() const {
    return version;
}

bool Dependency::prepare(const std::filesystem::path& source, const std::filesystem::path& destination) {
    std::smatch results;
    if (std::regex_search(path, results, repo)) {
        Project::info(path + " appears to be a git repository");
        const std::string& protocol = results[1];
        const std::string& host = results[2];
        const std::string& owner = results[3];
        const std::string& repo = results[4];

        return downloadRepo(destination, protocol, host, owner, repo);
    } else {
        Project::info("checking project at path " + path);
        std::filesystem::directory_entry srcDir(source / path);
        if (!srcDir.exists()) {
            Project::error("Project at " + path + " doesn't exist");
            return false;
        }
    }

    return true;
}

bool Dependency::downloadRepo(
    const std::filesystem::path& destination,
    const std::string& protocol,
    const std::string& host,
    const std::string& owner,
    const std::string& repo
) {
    nlohmann::json info;
    Project::info("Trying Gitea v1 API");
    bool res = repoGiteaApi1(protocol, host, owner, repo, info);

    if (res) {
        nlohmann::json::const_iterator itr = info.find("default_branch");
        if (itr != info.end() && itr->is_string()) {
            std::string branchName = *itr;
            Project::info("Gitea v1 API seem to have worked");
            Project::info("Default branch is " + branchName);

            std::string fileName = branchName + ".tar.gz";
            std::string url = protocol + "://" + host + "/api/v1/repos/" + owner + "/" + repo + "/archive/" + fileName;
            res = download(url, destination / "downloads" / fileName);
            if (res)
                Project::info("Successfully downloaded " + fileName);
        } else {
            res = false;
        }
    }

    if (!res) {
        Project::warn("Gitea v1 API didn't work");
    }

    return res;
}

bool Dependency::repoGiteaApi1(
    const std::string& protocol,
    const std::string& host,
    const std::string& owner,
    const std::string& repo,
    nlohmann::json& json
) {
    bool result = false;
    std::string url = protocol + "://" + host + "/api/v1/repos/" + owner + "/" + repo;
    std::string data;
    CURLcode code = httpGet(url, data, {acceptJson});
    if (code == CURLE_OK) {
        try {
            json = nlohmann::json::parse(data);
            result = true;
        } catch (const nlohmann::json::exception& e) {
            Project::warn(e.what());
        }
    }

    return result;
}

CURLcode Dependency::httpGet(const std::string& url, std::string& result, const std::list<std::string_view>& headers) {
    CURL* curl = curl_easy_init();
    curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
    curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1L);
    curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1L);

    struct curl_slist* curlHeaders = nullptr;
    for (const std::string_view& header : headers)
        curlHeaders = curl_slist_append(curlHeaders, header.data());

    curl_easy_setopt(curl, CURLOPT_HTTPHEADER, curlHeaders);
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, writeString);
    curl_easy_setopt(curl, CURLOPT_WRITEDATA, &result);

    curl_easy_setopt(curl, CURLOPT_VERBOSE, 1L);
    curl_easy_setopt(curl, CURLOPT_DEBUGFUNCTION, trace);
    curl_easy_setopt(curl, CURLOPT_DEBUGDATA, this);

    CURLcode res = curl_easy_perform(curl);

    curl_easy_cleanup(curl);
    return res;
}


bool Dependency::download(const std::string& url, const std::filesystem::path& destination) {
    std::filesystem::create_directories(destination.parent_path());
    if (std::filesystem::exists(destination))
        Project::minor("File " + destination.string() + " already exists, will be overwritten");

    bool result = false;
    CURL* curl = curl_easy_init();
    curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
    curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1L);
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, writeFile);

    curl_easy_setopt(curl, CURLOPT_VERBOSE, 1L);
    curl_easy_setopt(curl, CURLOPT_DEBUGFUNCTION, trace);
    curl_easy_setopt(curl, CURLOPT_DEBUGDATA, this);

    FILE* pagefile = fopen(destination.c_str(), "wb");
    if (pagefile) {
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, pagefile);
        CURLcode res = curl_easy_perform(curl);

        fclose(pagefile);
        if (res != CURLE_OK) {
            Project::warn("Couldn't download file " + url + ": " + curl_easy_strerror(res));
        } else {
            uint32_t code;
            curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &code);
            if (code == 200)
                result = true;
            else
                Project::warn("Couldn't download file " + url + ": response code " + std::to_string(code));
        }
    } else {
        Project::error(std::string("couldn't open output file ") + destination.c_str());
    }

    curl_easy_cleanup(curl);

    if (!result) {
        Project::minor("Removing " + destination.string() + " since the donwload failed");
        std::filesystem::remove(destination);
    }

    return result;
}

size_t Dependency::writeString(void* data, size_t size, size_t nmemb, void* pointer) {
    size_t finalSize = size * nmemb;
    std::string* string = static_cast<std::string*>(pointer);
    string->append(static_cast<char*>(data), finalSize);

    return finalSize;
}

size_t Dependency::writeFile(void* data, size_t size, size_t nmemb, void* file) {
    size_t written = fwrite(data, size, nmemb, (FILE *)file);
    return written;
}

int Dependency::trace(CURL* handle, curl_infotype type, char* data, size_t size, void* clientp) {
    switch (type) {
        case CURLINFO_TEXT: {
            std::string message(data, size);
            if (message[size - 1] == '\n')
                message = message.substr(0, size - 1);

            Project::debug(message);
        } break;
        default:
            break;
    }

    return 0;
}

